package org.lsst.gruth.grutils

/**
 * Base script class for Gru DSL.
 * <BR>
 *  Contains all the base calls (functions the describe a test set)
 *  and calls common to all contexts.
 *  <BR>
 *      There are three phase for tests:
 *      <UL>
 *          <LI> preparation: tests specifications are collected
 *          <LI> execution: tests are executed and test report collected
 *          <LI> reporting
 *       </UL>
 *  <BR>
 *   Base calls:
 *   <UL>
 *       <LI> <TT>_withClass</TT> specification of groups of tests for constructors, static factories,
 *       static methods.
 *       <LI> <TT>_withObjects</TT> specification of groups of tests for methods
 *   </UL>
 *   Base utilities functions ;
 *   <UL>
 *       <LI> <TT>_this</TT> the current script context (Binding object)
 *       <LI> <TT>_vars</TT> for declaring a set of "tagged" objects
 *       <LI> <TT>_runAll</TT> for requesting the execution af all pending tests (that have not yet been executed)
 *       reporting is also included since this is automatically executed at the end of each script.
 *   </UL>
 *
 *   (TODO: document _thisG, and _thisL -group and local binding-)
 *   <BR>
 *
 *   Functions and variables accessible only during a test run:
 *   <UL>
 *       <LI> assertions for the "expectations" phase (<TT>_xpect</TT> spec).
 *       <TT>_failIf, _failIfNot, _warnIf, _warnIfNot, _message, _neutral, _ofIfCaught
 *           </TT>
 *        <LI> read_only variables in the local context of a test run:
 *        some are always defined:
 *        <TT>
 *            _className, _methodName, _args, _result, _it
 *            </TT> (the current object under test)
 *            <BR>
 *         some are defined only in unit tests context (not in interception context):
 *         they may yield a null value.
 *         <TT>
 *            _rawTestName, _argsNames,
 *         </TT>
 *
 *   </UL>
 */

abstract class GruScript extends Script {
    //bizarre twist .....
    static GruScript running

    //TODO: use ServiceLoader to load ResultsHandlers (static code)

    /**
     * A current binding for all the script.
     * Typically variables declared through the <TT>_vars</TT>  are setted in the _this context.
     * A special variable <TT>NO_ARG</TT> is kept here.
     * It is also where "global" variables are kept.
     * <P>
     *  Global variables :
     *  <UL>
     *      <LI> <TT>currentGroup</TT>  set only when a group specification is running
     *      (<TT>_group</TT> method)
     *      <LI> <TT>currentRun</TT> the current <TT>TztRun</TT> object being executed.
     *      (important)
     *      <LI> <TT>groupNumber</TT> used to give a number when a group has no name specified
     *      <LI> <TT>groups</TT> all the groups that are not yet executed. this is a map were
     *      each group is classed along a <TT>SpecType</TT>.
     *      Normally they must be executed in a precise order according to this type (NOT YET IMPLEMENTED)
     *  </UL>
     */
    Binding _this = new Binding()
    /**
    *  the calling environment
    *  (not used for the moment)
    */
    Binding _env

    GruScript() {
        super()
        _env = getBinding()
        initThis()
        //setProperty('_result', this.&_result)
    }

    GruScript(Binding binding) {
        super(binding)
        //TODO : fix!
        _env = getBinding()
        initThis()
    }

    private void initThis() {
        running = this
        _this.currentRun = null
        _this.currentGroup = null
        _this.groupNumber = 1
        def groupMap = [:]
        SpecType.values().each {
            groupMap.put(it, [:])
        }
        //println groupMap
        _this.groups = groupMap
        //_env.groups = groupMap
        _this._NO_ARG = new TaggedObject('NO_ARG', [])
        //TODO: get the _scriptName name
        def scriptName = System.getProperty("script.name")
        //println _scriptName
         _defaultBundle(scriptName)

    }

    /**
    *  run all the test still registered in the global "groups" variable
     *  and run the reports.
     *  <BR>
     *   after this call the groups variable has no groups to run.
    * @return
    */
    def _runAll() {
        def map = _this.groups
        //println map
        // TODO: run in proper order: ctor, static factory, instance factory, instance methods, free
        // toDO: randomize method call order
        map.values().each {
            /*.println it
            it.keySet.each{
                println it
            }*/
            it.values().each {
                it.runReportFromGroupMap()
            }
        }
    }

    ////////////////////////////////////////////////////////// successive context
    //TODOS
    // create subclasses of GroupSpec for constructors, factories ansd methods
    // syntax
    //   _withClass <Class Spec> _group { <tests specs> }
    //   _withClass <Class Spec> _factory <static method> _group { }
    //   _withClass <Class Spec> _method <static method> _group { }
    //   _withObjects <list of NamedObjects/sets> _method <instance method> _group { }
    //   _withObjects <list of NamedObjects/sets> _factory <instance method> _group { }
    //   _group {}
    //  each _group could be of form _group groupName, {}


    GroupSpec currentBatchSpec;
    // TODO: synchronize! otherwise context problem

    /**
     * Starts registering a group of test specs for an existing class.
     * this suppose that the class exists when specifying the script.
     * @param clazz the class on which constructors or static code is to be tested
     * @return BuildGroupSpec object (see class)
     */
    BuildGroupSpec _withClass(Class clazz) {
        BuildGroupSpec batchSpec = new BuildGroupSpec(clazz.getCanonicalName(), _this)
        currentBatchSpec = batchSpec
        batchSpec.clazz = clazz
        MetaClass meta = clazz.metaClass

        batchSpec.execCode = meta.&invokeConstructor
        // sets up environment for pre, post and test code
        return batchSpec
    }

    /**
     * Enables a
     * simplified syntax for
     * <TT>_withClass clazz _group { /test code / } </TT>
     * that could be written :
     * <TT>_withClass class {/ test code / } </TT>
     * @param clazz Class on which to call constructor
     * @param clos group test specification
     * @return a group run
     */
    GroupRun _withClass(Class clazz, Closure clos) {
        return _withClass(clazz)._group (clos)
    }

    /**
    * Starts registering a group of test specs for a class that may exist of not.
     * if the class does not exist the results will be NOT_YET_IMPLEMENTED.
    *
    * @param className the class on which constructors or static code is to be tested
     * @return BuildGroupSpec object (see class)
    */
    BuildGroupSpec _withClass(String className) {
        BuildGroupSpec batchSpec = new BuildGroupSpec(className, _this)
        currentBatchSpec = batchSpec
        Class clazz = null;
        try {
            clazz = Class.forName className
            batchSpec.clazz = clazz
            MetaClass meta = clazz.metaClass
            batchSpec.execCode = meta.&invokeConstructor
        } catch (Throwable exc) {
            batchSpec.reportTemplate.caught << exc
            batchSpec.reportTemplate.rawDiagnostic = RawDiagnostic.NOT_YET_IMPLEMENTED
        }
        return batchSpec
    }

    BuildGroupSpec _withClass(String className, Closure closure) {
        return _withClass(className)._group (closure)

    }
    /**
    * Starts registering tests on tagged objects (instance methods are going to be tested)
    * @param objs
    * @return
    */
    //TODO: doc: explain what happens  if these objects have not been created.
    InvokeGroupSpec _withObjects(Object... objs) {
        InvokeGroupSpec batchSpec = new InvokeGroupSpec(_this, objs)
        currentBatchSpec = batchSpec
        return batchSpec
    }

    /**
     * Declares a set of Tagged objects. These are known to the local tests batch
     * (that is to the current gru file)
     *
     * @param map
     * @return
     */
    TaggedsMap _vars(Map map) {
        TaggedsMap res = new TaggedsMap(map)

        res.values().each {
            _this.setVariable(it.key, it)
        }

        return res

    }

    TaggedsMap _vars() {
        TaggedsMap res = [:]
        return res
    }

    /**
     * Helps resultReporters to determin a default bundle name for reports
     * (example: a base name for files containing the report)
     * @param name
     * @return
     */
    def _defaultBundle(String name) {
        ResultHandlers.reportBundleName = name
    }

    /**
     * same as _vars but these are going to be exported to other tests batch
     * <BR>
     *     NOT IMPLEMENTED
     * <BR>
     * TODO: with .gruh files?
     * @param map
     * @return
     */
    def _xport(String name, TaggedsMap map) {
        // if in .gruh don't reeexport

    }

    /**
    * NOT IMPLEMENTED
    * @param buildSpec
    * @return
    */
    // TODO: specs? change?
    def _run(BuildGroupSpec buildSpec) {
        // will expand and execute the tests and make the corresponding variables names
        // be in context of the script
        // and register in a dictionnary of variables
        // if BuildGroupSpec is incomplete -> error
        // the tests are not to be executed again
        // but results are known by gru
        // generate the tests request by combinations of parameters
        // general algorithm
        // these should be done once all the tests are gathered
        // randomize the order of execution ? no this is only at top level

        // execute each tests

        // pass the reports to reportHandler in alphabetical order
    }


    ////////////////////////////////////////////// _test method specs
    // all arguments are to be Gstring to be evaluated or boolean + string


    /**
     * Adds an <TT>AssertionReport</TT> to the current <TT>TztRun</TT> object:
     * the result is <TT>FAILED</TT> if the evaluation of the string is true
     * (the report also contains the evaluated string).
     * [NOT WORKING PROPERLY -YET-]
     * @param stringBooleanExpr a Gstring representing a valid Groovy expression
     * @return
     */
    def _failIf(String stringBooleanExpr) {
        if (_this.currentRun != null) {
            _this.currentRun._failIf(stringBooleanExpr)
        }

    }
    /**
     * Adds an <TT>AssertionReport</TT> to the current <TT>TztRun</TT> object:
     * the result is <TT>FAILED</TT> if the evaluation of the expression is true
     * (the report also contains the message)
     * @param message to be displayed in the report
     * @param booleanExpr before evaluating exceptionfired should be tested
     * @return
     */
    def _failIf(String message, boolean booleanExpr) {
        if (_this.currentRun != null) {
            _this.currentRun._failIf(message, booleanExpr)
        }


    }

    /**
    * same specs as failIf but fails if false
    * @param stringBooleanExpr
    * @return
    */
    def _failIfNot(String stringBooleanExpr) {
        if (_this.currentRun != null) {
            _this.currentRun._failIfNot(stringBooleanExpr)
        }

    }

    /**
     * same specs as failIf but fails if false
     */
    def _failIfNot(String message, boolean booleanExpr) {
        if (_this.currentRun != null) {
            _this.currentRun._failIfNot(message, booleanExpr)
        }

    }

    /**
     * Adds an <TT>AssertionReport</TT> to the current <TT>TztRun</TT> object:
     * the result is <TT>WARNINGS</TT> if the evaluation of the string is true
     * (the report also contains the evaluated string).
     * [NOT WORKING PROPERLY -YET-]
     * @param stringBooleanExpr a Gstring representing a valid Groovy expression
     * @return
     */
    def _warnIf(String stringBooleanExpr) {
        if (_this.currentRun != null) {
            _this.currentRun._warnIf(stringBooleanExpr)
        }

    }

    /**
     * Adds an <TT>AssertionReport</TT> to the current <TT>TztRun</TT> object:
     * the result is <TT>WARNINGS</TT> if the evaluation of the expression is true
     * (the report also contains the message)
     * @param message to be displayed in the report
     * @param booleanExpr
     * @return
     */
    def _warnIf(String message, boolean booleanExpr) {
        if (_this.currentRun != null) {
            _this.currentRun._warnIf(message, booleanExpr)
        }

    }

    /**
     * same specs as warnIf but warns if false
     */
    def _warnIfNot(String stringBooleanExpr) {
        if (_this.currentRun != null) {
            _this.currentRun._warnIfNot(stringBooleanExpr)
        }

    }

    /**
     * same specs as warnIf but warns if false
     */
    def _warnIfNot(String message, boolean booleanExpr) {
        if (_this.currentRun != null) {
            _this.currentRun._warnIfNot(message, booleanExpr)
        }

    }
    // do we need okIf and okIfNot ?
    /**
     * adds a message to the report
     * @return
     */
    def _message(String message) {
        if (_this.currentRun != null) {
            currentRun._message(message)
        }
        else {
            Cst._LOGGER.info(message)
        }
    }
       /**
    *  used to add some data to the report (does not change the result of the test)
    *
    * @param data any serializable data
    * @return the argument
    */
    Serializable _reportData(Serializable data) {
        if (_this.currentRun != null) {
            return currentRun._reportData(data)
        }
        else {
            Cst._LOGGER.info(message)
            return null
        }
    }

    /**
     * Adds an <TT>AssertionReport</TT> to the current <TT>TztRun</TT> object:
     * the result is <TT>NEUTRAL</TT> if the evaluation of the string is true
     * (the report also contains the evaluated string).
     * [NOT WORKING PROPERLY -YET-]
     * @param stringBooleanExpr a Gstring representing a valid Groovy expression
     * @return
     */
    def _neutral(String stringExpression) {
        if (_this.currentRun != null) {
            _this.currentRun._neutral(stringExpression)
        }

    }

    /**
     * Adds an <TT>AssertionReport</TT> to the current <TT>TztRun</TT> object:
     * the result is <TT>NEUTRAL</TT> if the evaluation of the expression is true
     * (the report also contains the message)
     * @param message to be displayed in the report
     * @param booleanExpr
     * @return
     */
    def _neutral(String message, Closure closure) {
        if (_this.currentRun != null) {
            _this.currentRun._neutral(message, closure)
        }

    }
    /**
     * Adds an <TT>AssertionReport</TT> to the current <TT>TztRun</TT> object:
     * forces the report to be ok if exception of this type is fires.
     * if any evaluation of assertions crashes
     * the result is still ok
     *
     * @arg  Throwable classes (should be a List?)
     * @return
     */
    def _okIfCaught(Class<? extends Throwable> throwClass) {
        // use the in operator to test if one exception is in Class
        if (_this.currentRun != null) {
            _this.currentRun._okIfCaught(throwClass)
        }
    }
    /**
     * normally not to be called but who knows?
     * @arg a list of Throwable classes
     * @return

     def _failIfCaught(Class<? extends Throwable> throwClass) {

     } */

    /// ACCESSORS TO current TestRUN
    String get_className() {
        if (_this.currentGroup != null) {
            return _this.currentGroup.className
        }
        if (_this.currentRun != null) {
            return _this.currentRun._className
        }
        throw new OutOfContextException("calling _className outside Group or Test")
    }

    String get_methodName() {
        if (_this.currentGroup != null) {
            return _this.currentGroup.methodName
        }
        if (_this.currentRun != null) {
            return _this.currentRun._methodName
        }
        throw new OutOfContextException("calling _methodName outside Group or Test")
    }
    // TODO: get_currentObject for method calls

    // TODO: implement a simple dynamic code for those accessors
    String get_rawTestName() {
        if (_this.currentRun != null) {
            return _this.currentRun._rootName
        }
        throw new OutOfContextException("calling _rawTestName outside Test run")
    }

    Object get_argsNames() {
        if (_this.currentRun != null) {
            return _this.currentRun._argsNames
        }
        throw new OutOfContextException("calling _argsNames outside Test run")
    }

    Object get_args() {
        if (_this.currentRun != null) {
            return _this.currentRun._args
        }
        throw new OutOfContextException("calling _args outside Test run")
    }

    Object get_result() {
        if (_this.currentRun != null) {
            return _this.currentRun._result
        }
        throw new OutOfContextException("calling _result outside Test run")
    }
    /*.
    void set_result(Object){
        throw new OutOfContextException("_result is read only")
    }
    */

    Object get_it() {
        if (_this.currentRun != null) {
            return _this.currentRun._it
        }
        throw new OutOfContextException("calling _it outside Test run")
    }

    Object get_currentObject() {
        if (_this.currentRun != null) {
            return _this.currentRun.get_currentObject()
        }
        throw new OutOfContextException("calling _currentObject outside Test run")

    }
    Object get_currentKey() {
        if (_this.currentRun != null) {
            return _this.currentRun.get_currentKey()
        }
        throw new OutOfContextException("calling _currentKey outside Test run")
    }


        // TODO: beware! may be evaluated at group time : change!
    Binding get_thisL() {
        if (_this.currentRun != null) {
            return _this.currentRun._thisL
        }
        throw new OutOfContextException("calling local context outside Test run")
    }


}
