package org.lsst.gruth.grutils

import java.util.logging.Level

/**
 *  Reports about the execution of a test.
 *  this structure is to be kept , stored and may be modified by other software
 * @author bamade
 *
 */
class TztReport implements Cloneable, WithStringKey, Serializable,  Comparable<TztReport>  {
    /**
     * name of test : see _rootName in TztRun
     */
    String testName;
    RawDiagnostic rawDiagnostic = RawDiagnostic.NEUTRAL;
    String methodName;
    String className;
    // azzertions array
    List<String> messages = [];
    /**
     * this Object may not be serialised: TaggedObject takes care of that
     */
    TaggedObject supportingObject;

    /**
     * Object produced by test ... if not serializable
     * will be dumped at Serialization. (or replaced by a String)
     */
    // TODO: URGENT write serialization methods
    transient Object data;
    /**
     * the result object does not have an "equals" method that can be used after serialization/deserialisation
     */
    boolean noEqualsForResult = false ;

    /**
     * used for reporting anything
     */
    Serializable additionalData;

    List<Throwable> caught = [];
    List<AssertionReport> listAssertions = []

    boolean exceptionFired = false
    // Advice advice = Advice.NOT_EVALUATED_YET; // TODO: decide whether o not to initialize the value here

    /**
     * used to declare this run not to be included in final report
     */
    transient boolean dontReport = false

    transient Binding _this = GruScript.running?._this

    TztReport(String className, String methodName) {
        this.className = className
        this.methodName = methodName
    }

    TztReport(String className) {
        this.className = className
    }

    TztReport() {

    }

    // for test purposes only
    TztReport(String testName, RawDiagnostic rawDiagnostic,  Object data, List<Throwable> caught) {
        this.testName = testName
        this.rawDiagnostic = rawDiagnostic
        if(caught!= null) this.caught = caught
        this.data = data
        this.className = ""
        this.methodName = ""
    }

    private void writeObject(java.io.ObjectOutputStream out) throws IOException {
        out.defaultWriteObject()
        if(! (this.data instanceof Serializable)) {
            out.writeObject('-not serializable-:'+this.data.toString())
        } else {
            out.writeObject( this.data)
        }
    }

    private void readObject(java.io.ObjectInputStream input) throws IOException, ClassNotFoundException {
        input.defaultReadObject()
        this.data = input.readObject()
    }


    TztReport clone() {
        TztReport res = (TztReport) super.clone()
        if(caught != null) res.caught = caught.clone()
        if(messages != null) res.messages = messages.clone()
        if(listAssertions != null) res.listAssertions = listAssertions.clone()
        return res;
    }

    public String toString() {
        return """class: ${this.className} method: ${this.methodName} message [${this.messages}]
        rawDiagnostic: ${this.rawDiagnostic} caughts: ${this.caught}""".toString()
    }

    public String getSupportingObjectString() {
        Object supporting;
        if(this.supportingObject instanceof TaggedObject) {
            supporting = ((TaggedObject)this.supportingObject).key
        } else {
            supporting = this.supportingObject
        }
        if(supporting == null) {
            supporting = className ;
        }
        return String.valueOf(supporting)
    }

    String getKey() {
        return this.testName
    }

    public String getUniqueKey(){
        return getKey() + "\u21E8" + getSupportingObjectString()
    }
    ////// ASSERTIONS HANDLING

    /**
     * Adds an <TT>AssertionReport</TT> to the current  object:
     * the result is <TT>FAILED</TT> if the evaluation of the string is true
     * (the report also contains the evaluated string).
     * [NOT WORKING PROPERLY -YET-]
     * @param stringBooleanExpr a Gstring representing a valid Groovy expression
     * @return
     */
    def _failIf(String stringBooleanExpr) {
        registerAssertion(stringBooleanExpr, true, RawDiagnostic.FAILED)
    }
    /**
     * Adds an <TT>AssertionReport</TT> to the current  object:
     * the result is <TT>FAILED</TT> if the evaluation of the expression is true
     * (the report also contains the message)
     * @param message to be displayed in the report
     * @param booleanExpr before evaluating exceptionfired should be tested
     * @return
     */
    def _failIf(String message, boolean booleanExpr) {
        registerAssertion(message, booleanExpr, true, RawDiagnostic.FAILED)
    }

    /**
     * same specs as failIf but faisl if false
     * @param stringBooleanExpr
     * @return
     */
    def _failIfNot(String stringBooleanExpr) {
        registerAssertion(stringBooleanExpr, false, RawDiagnostic.FAILED)
    }

    /**
     * same specs as failIf but fails if false
     */
    def _failIfNot(String message, boolean booleanExpr) {
        registerAssertion(message, booleanExpr, false, RawDiagnostic.FAILED)
    }

    /**
     * Adds an <TT>AssertionReport</TT> to the current object:
     * the result is <TT>WARNINGS</TT> if the evaluation of the string is true
     * (the report also contains the evaluated string).
     * [NOT WORKING PROPERLY -YET-]
     * @param stringBooleanExpr a Gstring representing a valid Groovy expression
     * @return
     */
    def _warnIf(String stringBooleanExpr) {
        registerAssertion(stringBooleanExpr, true, RawDiagnostic.WARNINGS)
    }

    /**
     * Adds an <TT>AssertionReport</TT> to the current  object:
     * the result is <TT>WARNINGS</TT> if the evaluation of the expression is true
     * (the report also contains the message)
     * @param message to be displayed in the report
     * @param booleanExpr
     * @return
     */
    def _warnIf(String message, boolean booleanExpr) {
        registerAssertion(message, booleanExpr, true, RawDiagnostic.WARNINGS)
    }

    /**
     * same specs as warnIf but warns if false
     */
    def _warnIfNot(String stringBooleanExpr) {
        registerAssertion(stringBooleanExpr, false, RawDiagnostic.WARNINGS)
    }

    /**
     * same specs as warnIf but warns if false
     */
    def _warnIfNot(String message, boolean booleanExpr) {
        registerAssertion(message, booleanExpr, false, RawDiagnostic.WARNINGS)
    }

    public void _message(String message) {
        this.messages << message
    }

    /**
    *  used to add some data to the report (does not change the result of the test)
    *
    * @param data any serializable data
    * @return the argument
    */
    Serializable _reportData(Serializable data) {
        this.additionalData = data
        return data
    }

    def _neutral(String stringExpression) {
        try {
            //TODO; does not work!
            def res = groovy.util.Eval.me(stringExpression)
            registerAssertion("${stringExpression} -> ${res}", true, true, RawDiagnostic.NEUTRAL)
        } catch (Exception exc) {
            reportException(exc, RawDiagnostic.TOOL_OR_CODE_ERROR)
        }

    }

    def _neutral(String message, Closure closure) {
        try {
            def res = closure.call()
            registerAssertion("${message} -> ${res}", true, true, RawDiagnostic.NEUTRAL)
        } catch (Exception exc) {
            reportException(exc, RawDiagnostic.TOOL_OR_CODE_ERROR)
        }

    }

    def _doNotReport() {
        this.dontReport= true
    }

    def _resultHasNoEquals() {
        this.noEqualsForResult = true
    }

    def _okIfCaught(Class<? extends Throwable> throwClass) {
        // use the in operator to test if one exception is in Class
        boolean found = false
        this.caught.each {
            if (it in throwClass) {
                listAssertions << new AssertionReport(true, it.toString(), RawDiagnostic.SUCCESS)
                found = true
            }
        }
        if (!found) {
            listAssertions << new AssertionReport(true, "not found ${throwClass}", RawDiagnostic.FAILED)
        }
    }

    protected void registerAssertion(String stringBooleanExpression, boolean expected, RawDiagnostic result) {
        try {
            throw new UnsupportedOperationException("Groovy bug on dynamic operations?")
            //TODO; does not work! when invoking -> missing property
            // create a Stream and have it evaluated by the current script
            boolean res = groovy.util.Eval.me(stringBooleanExpression)
            //boolean res = _this.currentScript.evaluate(stringBooleanExpression)
            registerAssertion(stringBooleanExpression, res, expected, result)
        } catch (Exception exc) {
            reportException(exc, RawDiagnostic.TOOL_OR_CODE_ERROR)
        }

    }

    protected void registerAssertion(String message, boolean actual, boolean expected, RawDiagnostic result) {
        AssertionReport assertion
        if (actual == expected) {
            assertion = new AssertionReport(message, result)
        } else {
            assertion = new AssertionReport(message, RawDiagnostic.NEUTRAL)
        }
        listAssertions << assertion
    }

    void reportException(Throwable th, RawDiagnostic result) {
        this.caught << th
        this.rawDiagnostic = result
        this.exceptionFired = true
        // TODO: problem with init of Cst!!!!
        Cst._LOGGER.log(Level.FINE, "", th)
    }

    void gatherResult() {
         // now goes through the exceptions
            // if < Ok_MARK takes the worst
            // if > OK_MARK takes the worst ?
            // if forced takes the value
            RawDiagnostic res = this.rawDiagnostic
            boolean forced = false
            this.listAssertions.each {
                RawDiagnostic nextResult = it.result
                if (it.force) {
                    res = nextResult
                    forced = true;
                }
                if ((!forced) && (nextResult < res)) {
                    if (nextResult != RawDiagnostic.NEUTRAL) {
                        res = nextResult
                    }
                }

            }
            this.rawDiagnostic = res;
    }

    def _fill(TaggedsMap tMap) {
        if ( (! exceptionFired) && (rawDiagnostic >= RawDiagnostic.WARNINGS )) {
           String name = this.testName
            /* TODO: CHANGE TO
            *
             */
           if(name == null) {
               tMap.put(this.data)
           } else {
               tMap.put(name, data)
           }
            //tMap.put(this.data)
        }
    }

    @Override
    int compareTo(TztReport other) {
        return (this.getUniqueKey()).compareTo(other.getUniqueKey());
    }
}
