package org.lsst.gruth.ggre;


import org.lsst.gruth.grutils.Advice;
import org.lsst.gruth.grutils.AnnotatedReport;
import org.lsst.gruth.grutils.TztReport;

/**
 * This class defines the structure to be used by an {@link java.util.Observable Observable}
 * to {@link java.util.Observable#notifyObservers(Object) push-notify}
 * its {@link java.util.Observer Observer}(s) about
 * <ul>
 * 		<li>a selected {@link TztReport report} </li>
 * 		<li>and its state (edited or not).</li>
 * </ul>
 *
 * @author Alain BECKER <alain.becker@gmail.com>
 *
 */
public class ReportsPanelNotification {



    private final AnnotatedReport originalReport;
    /**
     * @return the original {@link TztReport Report}
     */
    public AnnotatedReport getOriginalReport() {
        return originalReport;
    }



    private final AnnotatedReport editedReport;
    /**
     * @return the {@link TztReport Report} (don't presume it has <em>actually</em> been edited;
     * 			use <code>isDifferenceNotification()</code> to get a clue)
     * @see ReportsPanelNotification#isDifferenceNotification()
     */
    public AnnotatedReport getEditedReport() {
        return editedReport;
    }


    /**
     * @return The report index, if any (else, <code>null</code>)
     */
    private final int index;
    public int getIndex() {
        return this.index;
    }



    private final boolean differenceNotification;
    /**
     * @return a boolean telling (if <code>true</code>) that the two {@link TztReport Report}s differ
     * 			(or that the {@link java.util.Observable} thought they should be considered as different).
     */
    public boolean isDifferenceNotification() {
        return differenceNotification;
    }



    /**
     * Use this constructor if you <strong>can</strong> tell that the reports differ,
     * or if you want you {@link java.util.Observer}(s) to believe so.
     *
     * <p>Another constructor exists to compute the difference flag.</p>
     * @see #ReportsPanelNotification(TztReport, TztReport)
     *
     * @param originalReport The original, unedited, {@link TztReport Report}
     * @param editedReport The current state of the (potentially edited) {@link TztReport Report}
     * @param index The report index, if any (else, <code>null</code>)
     * @param differenceNotification <code>true</code> to tell the {@link java.util.Observer}(s)
     * 			that the two {@link TztReport Report}s contain different values.
     * 			Use <code>null</code> for the difference to be computed.
     */
    protected ReportsPanelNotification(AnnotatedReport originalReport, AnnotatedReport editedReport,  int index, Boolean differenceNotification) {
        if (differenceNotification == null) {
            //differenceNotification = ! originalReport.getAdvice().equals(editedReport.getAdvice()); // compute difference
            // differenceNotification =  (originalReport.getAdvice() == null) ^ (editedReport.getAdvice() == null); // compute difference
            if (originalReport.getAdvice() == null) {
                differenceNotification = editedReport.getAdvice() != null;
            } else {
                differenceNotification = ! originalReport.getAdvice().equals(editedReport.getAdvice());
            }
        }
        this.originalReport = originalReport;
        this.editedReport = editedReport;
        this.index = index;
        this.differenceNotification = differenceNotification.booleanValue();
    }

    /**
     * Use this constructor if you want it to compute reports
     * {@link ReportsPanelNotification#isDifferenceNotification() difference} for you.
     *
     * <p>Another constructor exists to manually set the difference flag.</p>
     * @see #ReportsPanelNotification(TztReport, TztReport, boolean)
     *
     * @param originalReport The original, unedited, {@link TztReport Report}
     * @param editedReport The current state of the (potentially edited) {@link TztReport Report}
     * @param index The report index, if any (else, <code>null</code>)
     */
    public ReportsPanelNotification(AnnotatedReport originalReport, AnnotatedReport editedReport, int index) {
        this( originalReport, editedReport, index, null);
    }


}
