package org.lsst.gruth.jutils;

import groovy.lang.Range;

import java.io.Serializable;
import java.lang.reflect.Array;
import java.math.BigDecimal;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

/**
 * (could not re-use java.lang.Reference)
 * A Placeholder for a parameter value (for constructor or method).
 * Such an object has a value that an be modified : it can be empty (then method <TT>isSet</TT>
 * yields false)  or initialized.
 * <P>
 *     A special feature is about null values: normally setting up the value with null
 *     makes it uninitialized so making it nul AND initializes should be performed
 *     either by  <TT>setNullValue</TT> or through  <TT>modifyChecked</TT>
 *     which is a generic way of setting a value (with type control).
 * </P>
 *
 * @author bamade
 */
public  abstract class HollowParm<T extends Serializable>  implements Serializable, Cloneable{
    protected T value ;
    protected boolean set ;
    protected boolean readOnly = false ;
    protected Properties properties ;

    protected HollowParm() {}
    protected HollowParm(T value) {
        this.value = value ;
        if(value != null) {
            set = true ;
        }
    }

    protected HollowParm(T value, Map<?,?> map){
        this(value) ;
        setProperties(map);
    }

    final void setProperties(Map<?,?> map) {
        properties = new Properties() ;
        for(Map.Entry<?,?> entry : map.entrySet()){
            Object valProp = entry.getValue() ;
            String valPropStr = stringify(valProp);
            if(valProp instanceof BigDecimal) {
                valPropStr = ((BigDecimal)valProp).toPlainString() ;
            } else {
                valPropStr = String.valueOf(valProp) ;
            }
            properties.setProperty(entry.getKey().toString(), valPropStr) ;
        }
    }

    public static String stringify(Object obj){
        String res ;
        if(obj instanceof BigDecimal) {
            res = ((BigDecimal)obj).toPlainString() ;
        }  else if (obj instanceof Range) {
            Range range = (Range) obj ;
            Object lower = range.getFrom();
            Object upper = range.getTo();
            res = stringify(lower)+ ".." + stringify(upper) ;
        }else if (obj instanceof Map) {
            // Dos not work properly : res= DefaultGroovyMethods.toMapString((Map) obj);
            res= toMapString((Map) obj);
        } else if (obj instanceof List) {
            // does not work properly : res= DefaultGroovyMethods.toListString((List) obj);
            res= toListString((List) obj);
        } else if(obj.getClass().isArray()) {
            res =  toArrayString(obj) ;
        }
        else {
            res = String.valueOf(obj) ;
        }
        return res ;

    }

    private static String  toArrayString(Object obj){
        StringBuilder builder = new StringBuilder("[");
        int size = Array.getLength(obj) ;
        for(int ix = 0 ; ix < size ; ix++) {
            Object item = Array.get(obj, ix);
            if(item instanceof String) {
                builder.append("'").append(item).append("'") ;
            } else {
                builder.append(stringify(item)) ;
            }

            if(ix < (size -1)) {
                builder.append(",");
            }
        }
        builder.append("]") ;
        return builder.toString();
    }
    //TODO : Double and Float string changed to "as double" "as Float" ?
    // rationale: this is of no importance for constuctors
    // but will probably need a distinct behaviour when we use "raw" lists
    private static String  toListString(List list){
        StringBuilder builder = new StringBuilder("[");
        int size = list.size();
        for(int ix = 0 ; ix < size ; ix++) {
            Object item = list.get(ix);
            if(item instanceof String) {
                builder.append("'").append(item).append("'") ;
            } else {
                builder.append(stringify(item)) ;
            }

            if(ix < (size -1)) {
                builder.append(",");
            }
        }
        builder.append("]") ;
        return builder.toString();
    }
    //TODO :change double and floats values as for List?
    // rationale: this is of no importance for constuctors
    // but will probably need a distinct behaviour when we use "raw" lists
    private static String  toMapString(Map map){
        StringBuilder builder = new StringBuilder("[");
        Set<Map.Entry> entries = map.entrySet() ;
        int size =entries.size() ;
        int ix = 0 ;
        for(Map.Entry entry : entries) {
            builder.append(entry.getKey()).append(":") ;
            Object item = entry.getValue() ;
            if(item instanceof String) {
                builder.append("'").append(item).append("'") ;
            } else {
                builder.append(stringify(item)) ;
            }
            if(ix < (size -1)) {
                builder.append(",");
            }
            ix++ ;
        }
        builder.append("]") ;
        return builder.toString();
    }

    public Properties getProperties() {
        return properties;
    }

    public T getValue() {
        return value;
    }

    public void setValue(T value) {
        if (readOnly) return ;
        this.value = value;
        if(value != null) {
            set = true ;
        } else {
            set = false ;
        }
    }

    public void setNullValue() {
        if (readOnly) return ;
        this.value = null ;
        set = true ;
    }

    public void clear() {
        if (readOnly) return ;
        this.setValue(null);
    }

    public void modifyChecked(Object obj) {
        if (readOnly) return ;
        if(obj == null) {
            setNullValue();
            return ;
        }
        Class clazz = obj.getClass() ;
        Class myClass = getValueClass() ;
        if(! myClass.isAssignableFrom(clazz)) {
            throw new  IncompatibleTypeException(myClass, clazz) ;
        }
        this.value = (T)obj ;
    }


    public boolean isSet() {
        return set;
    }

    public boolean isReadOnly() {
        return readOnly;
    }

    public String toString() {
        //TODO: check why changing this provokes a bug!!
        //return this.getClass().getSimpleName() + " " + String.valueOf(value) ;
        //return  String.valueOf(value) ;
        return stringify(value) ;
    }
    public abstract Class<T> getValueClass() ;

    public HollowParm<T> addProperties(Map.Entry<String, String>... entries){
        if(this.properties == null) {
            this.properties = new Properties() ;
        }
        for(Map.Entry<String, String> entry : entries){
            properties.setProperty(entry.getKey(), entry.getValue()) ;
        }
        return this ;
    }

    /**
     * first level clone: object that are referenced are not duplicated
     * @return
     */
    public  HollowParm clone() {
        HollowParm nextOne = null ;
        try {
            nextOne = (HollowParm) super.clone() ;
            if(properties!= null){
                nextOne.properties = (Properties) properties.clone() ;
            }
        } catch (CloneNotSupportedException e) {
            /* should not happen */
        }
        return nextOne ;
    }

}
